package geniusweb.boa.acceptancestrategy;

import java.util.HashMap;

import geniusweb.bidspace.BidsWithUtility;
import geniusweb.boa.BoaState;
import geniusweb.boa.biddingstrategy.TimeDependentBiddingStrategy;
import geniusweb.inform.Settings;
import geniusweb.issuevalue.Bid;
import geniusweb.profile.Profile;
import geniusweb.profile.utilityspace.LinearAdditive;
import geniusweb.profile.utilityspace.UtilitySpace;

/**
 * Simple implementation similar to the {@link TimeDependentBiddingStrategy}.
 * Uses 4 parameters:
 * <ul>
 * <li>e: the exponent for the target-utility function. See {@link #f(double)}
 * <li>k: the constant for the target-utility function. See {@link #f(double)}
 * <li>max: the maximum utility target used at t=0.
 * <li>min: the minimum utility target used at t=1
 * </ul>
 */
public class TimeDependentAcceptanceStrategy implements AcceptanceStrategy {

	private Double min = null, max = null, e = null, k = null;

	@Override
	public Boolean isAcceptable(Bid bid, BoaState state) {
		if (min == null || max == null || e == null || k == null) {
			min = getMin(state);
			max = getMax(state);
			e = getE(state);
			k = getK(state);
		}
		double targetUtil = p(
				state.getProgress().get(System.currentTimeMillis()));
		// we subtract epsilon because of rounding errors in computation.
		return ((UtilitySpace) state.getProfile()).getUtility(bid)
				.doubleValue() >= targetUtil - 0.0000001;
	}

	/**
	 * Overrideable for hard configuring this component.
	 * 
	 * @param state the {@link BoaState}
	 * @return the parameter e for the time depemdency function
	 *         {@link #f(double)}. The parameter is 1 by default, or the value
	 *         for the "e" parameter in the {@link Settings} if available.
	 */
	protected Double getE(BoaState state) {
		HashMap<String, Object> params = state.getSettings().getParameters();
		Double e = 1d;
		if (params.containsKey("e")) {
			e = (Double) params.get("e");
			if (e < 0 || e > 1)
				throw new IllegalArgumentException(
						"e must be in [0,1] but got " + this.e);
		}
		return e;
	}

	/**
	 * Overrideable for hard configuring this component.
	 * 
	 * @param state the {@link BoaState}
	 * @return the parameter k for the time depemdency function
	 *         {@link #f(double)}. The parameter is 0 by default, or the value
	 *         for the "k" parameter in the {@link Settings} if available.
	 */
	protected Double getK(BoaState state) {
		HashMap<String, Object> params = state.getSettings().getParameters();
		Double k = 0d; // default
		if (params.containsKey("k")) {
			k = (Double) params.get("k");
			if (k < 0 || k > 1)
				throw new IllegalArgumentException(
						"k must be in [0,1] but got " + this.e);
		}
		return k;
	}

	/**
	 * Overrideable for hard configuring this component.
	 * 
	 * @param state the {@link BoaState}
	 * @return the min value for {@link #p(double)}. We use the "min" parameter
	 *         in the {@link Settings} if available. If not available, the
	 *         parameter is computed as the utility of the reservation bid. If
	 *         there is no reservation bid, we use the minimum utility of the
	 *         available profile.
	 */
	protected Double getMin(BoaState state) {
		HashMap<String, Object> params = state.getSettings().getParameters();
		if (params.containsKey("min")) {
			return Math.max(0f, Math.min(1f, (Double) params.get("min")));
		}
		UtilitySpace profile = (UtilitySpace) state.getProfile();
		if (profile.getReservationBid() != null) {
			return profile.getUtility(profile.getReservationBid())
					.doubleValue();
		}

		return getBidspace(profile).getRange().getMin().doubleValue();
	}

	/**
	 * Overrideable for hard configuring this component.
	 * 
	 * @param state the {@link BoaState}
	 * @return the max value for {@link #p(double)}. We use the "max" parameter
	 *         in the {@link Settings} if available. If not available, we use
	 *         the maximum utility of the available profile.
	 */
	protected Double getMax(BoaState state) {
		HashMap<String, Object> params = state.getSettings().getParameters();
		if (params.containsKey("max")) {
			return Math.max(0f, Math.min(1f, (Double) params.get("max")));
		}
		UtilitySpace profile = (UtilitySpace) state.getProfile();
		return getBidspace(profile).getRange().getMax().doubleValue();

	}

	/**
	 * factory method. Overridable for testing
	 * 
	 * @param profile the profile to get a bidspace for
	 * @return {@link BidsWithUtility}
	 */
	protected BidsWithUtility getBidspace(Profile profile) {
		return new BidsWithUtility((LinearAdditive) profile);
	}

	/**
	 * Makes sure the target utility with in the acceptable range according to
	 * the domain
	 * 
	 * @param t the normalized current time in the negotiation, where t=9 at
	 *          start of negotiation and t=1 at end of negotiation.
	 * @return double
	 */
	private double p(double t) {
		return this.min + (this.max - this.min) * (1.0 - f(t));
	}

	/**
	 * From the paper:
	 * 
	 * A wide range of time dependent functions can be defined by varying the
	 * way in which f(t) is computed. However, functions must ensure that 0 <=
	 * f(t) <= 1, f(0) = k, and f(1) = 1.
	 * 
	 * That is, the offer will always be between the value range, at the
	 * beginning it will give the initial constant and when the deadline is
	 * reached, it will offer the reservation value.
	 * 
	 * @param t the normalized current time in the negotiation, where t=9 at
	 *          start of negotiation and t=1 at end of negotiation.
	 * @return
	 */
	private double f(double t) {
		return k + (1.0 - k) * Math.pow(t, 1.0 / e);
	}

}